/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
package com.openexchange.office.tools.user;

import java.util.HashMap;

import org.apache.commons.lang3.Validate;

import com.openexchange.realtime.packet.ID;

/**
 * Manages the user data for every member of a connection instance.
 *
 * @author Carsten Driesner
 * @since 7.8.0
 */
public class UserDataManager {

    private final HashMap<String, UserData> userDataMap = new HashMap<String, UserData>();
    private UserData lastUserData = null;
    private String lastUserKey = null;

    /**
     * Initalizes a new UserDataManager instance.
     */
    public UserDataManager() {

    }

    /**
     * Determines if user-specific data is associated with the real-time id.
     *
     * @param id the real-time id of the user.
     * @return true if user data is available and false if not
     */
    public synchronized boolean hasUserData(final ID id) {
        Validate.notNull(id);

        final String key = id.toString();
        return userDataMap.containsKey(key) || key.equalsIgnoreCase(lastUserKey);
    }

    /**
     * Retrieves the user-specific data associated with the real-time id.
     *
     * @param id the real-time id of the user.
     * @return the user-specific data or null
     */
    public synchronized final UserData getUserData(final ID id) {
        Validate.notNull(id);

        final String key = id.toString();
        if (!userDataMap.isEmpty()) {
            return userDataMap.get(key);
        } else if (key.equalsIgnoreCase(lastUserKey)) {
            return lastUserData;
        }

        return null;
    }

    /**
     * Stores the user-specific data associated with the real-time id.
     *
     * @param id the real-time id of the user.
     * @param userData the user-specific data to be stored in association with the real-time id
     */
    public synchronized void addUser(final ID id, final UserData userData) {
        Validate.notNull(id);

        userDataMap.put(id.toString(), userData);
    }

    /**
     * Removes the user-specific data associated with a real-time id.
     *
     * @param id the real-time id where the user-specific data should be removed.
     */
    public synchronized void removeUser(final ID id) {
        Validate.notNull(id);

        final UserData userData = userDataMap.remove(id.toString());
        if (userDataMap.isEmpty()) {
            // store last user data for use after the client left, e.g. dispose
            lastUserData = userData;
            lastUserKey = id.toString();
        }
    }

}
